--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
-- ]]
--
-- [[
-- This module defines Succubus' powers. The bulk of Succubus' powers revolve
-- around redirection and debuffs. She only has a single attack power: Crop whip
-- which does damage and inflicts the Spanked ailment.
--
-- Most of her powers instead revolve around controlling or otherwise
-- inhibiting her opponents. However, these powers have a critical weakness:
-- they only really work when Succubus' power is greater than her opponent's.
-- Many of them are also unable to be used if Succubus has been spanked
-- recently.
-- ]]
--
-- [[
-- Contains some useful utility functions for working with powers, such as
-- dynamic gender functions.
-- ]]

local combatstateops = require 'combat/combatstate'
local log = require 'log'
local pages = require 'pages'
local powers = require 'combat/powers'
local scarlettables = require 'scarlettables'

local M = {}

function M.hisher(character)
    return character.attributes.gender == "female" and "her" or "his"
end

function M.HisHer(character)
    return character.attributes.gender == "female" and "Her" or "His"
end

function M.himher(character)
    return character.attributes.gender == "female" and "her" or "his"
end

function M.HimHer(character)
    return character.attributes.gender == "female" and "Her" or "His"
end


function M.himselfherself(character)
    return character.attributes.gender == "female" and "herself" or "himself"
end

function M.heshe(character)
    return character.attributes.gender == "female" and "she" or "he"
end

function M.HeShe(character)
    return character.attributes.gender == "female" and "She" or "He"
end

function M.heroheroine(character)
    return character.attributes.gender == "female" and "heroine" or "hero"
end

function M.lowerrisque(character, risque, notrisque)
    return character.clothing.lower.risque >= 4 and risque or notrisque
end

function M.baring(character)
    return character.clothing.lower.risque >= 5
end


-- [[
-- Applies the cost of the specified power to the specified character.
--
-- combatstate: The current state of combat
-- power: The power to apply the cost of
-- ]]
function M.inflictcost(combatstate, power)
    local activecharacter = combatstateops.activecharacter(combatstate)
    powers.applycost(activecharacter, power)
end

-- [[
-- Given a table of characters, returns the character with the median energy.
-- If the number of characters is even, then we alternate between the character
-- with the median energy, and the character just after them in the list. This
-- is primarily so that powers that target a median energy character don't just
-- hit the lowest every time. We use the current round to drive when we use
-- the "median" character or the character with the next highest energy.
--
-- characters: The list of characters to find the median energy for
-- round: The current round of combat
--
-- return The character with the median energy
-- ]]
function M.median(characters, round)
    local characterenergies = {}
    for _ ,character in pairs(characters) do
        table.insert(characterenergies, character.statistics.energy)
    end
    table.sort(characterenergies)
    local modifier = #characterenergies % 2 == 0 and round % 2 or 0
    local medianindex = math.ceil(#characterenergies / 2) + modifier
    local medianenergy = characterenergies[medianindex]
    for _, character in pairs(characters) do
        if character.statistics.energy == medianenergy then
            return character
        end
    end
end

--[[
-- Cuts the specified character's energy in half and creates a duplicate of the
-- character with the specified name
--
-- Returns false if the summon doesn't happen, true if it does
--]]
function M.summon(combatstate, character, summonname, summonconcisename)
    if M.issummon(character) then
        return false
    end
    local statistics = character.statistics
    for _, otherchar in pairs(combatstate.characters) do
        if otherchar.combatname == summonname then
            if combatstateops.isdefeated(combatstate, otherchar.name) then
                otherchar.statistics.energy = math.floor(statistics.energy / 2)
                combatstateops.revivecharacter(combatstate, otherchar.name)
                character.statistics.energy = math.floor(statistics.energy / 2)
                return true
            else
                return false
            end
        end
    end
    character.statistics.energy = math.floor(statistics.energy / 2)
    local summon = scarlettables.copy(character)
    summon.combatname = summonname
    summon.printedname = summonname
    summon.concisename = summonconcisename
    summon.name = character.name .. "_summon"
    if combatstateops.isally(combatstate, character) then
        combatstateops.addally(combatstate, summon)
    else
        combatstateops.addenemy(combatstate, summon)
    end
    return true
end

function M.issummon(character)
        return string.find(character.name, "_summon")
end

-- [[
-- Returns the summon for the specified character, if one already exists, nil otherwise.
-- Always returns true for summoned characters (so that they can't themselves summon
-- characters).
--]]
function M.findsummon(combatstate, character)
    return
        string.find(character.name, "_summon") or
        combatstateops.findcharacter(combatstate, character.name .. "_summon")
end

--[[
-- Given a power with all its metadata already defined, and a
-- function on combatstate that returns a combat primitive,
-- transforms the power into a combat page
--
-- powermetadata: A table defining the power's metadata. See
--  'combat/powers.md' for details.
-- primitivesource: A function from combatstate to combat primitive
--  implementing a power
--
-- returns The power with the specified metadata and primitives
--]]
function M.buildpower(powermetadata, primitivesource)
    local powerpage = pages.page(
        nil,
        primitivesource,
        nil,
        function(combatstate)
            M.inflictcost(combatstate, powermetadata)
        end
    )
    for key, value in pairs(powermetadata) do
        powerpage[key] = value
    end
    return powerpage
end

return M
