--[[
-- Copyright 2019 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--[[
-- This module defines Princess' powers. Princess is not at all a
-- fighter. However, she is capable of summoning, buffing, healing
-- and ordering toys about. These toys make her very dangerous.
--
-- This module has the following tables in it:
-- princess - Defines Princess' powers.
-- mrcuddles - Definees Mr. Cuddles's (the giant clay teddy bear Princess rides) powers
-- stuffins - Defines the powers for Sir Stuffins, the teddy bear wielding a pool noodle
-- giraffalot - Defines the powers for Giraffalot, Sir Stuffins' stead.
--]]

local M = {}

local attacks = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local combatutils = require 'combat/combatutils'
local F = require 'f-strings/F'
local log = require 'log'
local pages = require 'pages'
local powers = require 'combat/powers'
local powerutils = require 'data/powerutils'
local scarlettables = require 'scarlettables'
local statuses = require 'data/statuses'
local statusops = require 'combat/statuses'

local summoncount = 1
M.princess = {}

--[[
-- Summon toys summons enough toys (disposable minions) to max out the number
-- of enemies (9) that Scarlet Moon needs to fight.
--]]
M.princess.summontoys = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local characters = require 'data/characters'
        local character = combatstateops.activecharacter(combatstate)
        local numallies = #(combatstateops.activeallies(combatstate))
        local numsummons = math.max(0, 9 - numallies)
        local buff = math.floor(attacks.statistic(character, "power") / 4)
        local msg = 
            "\"Come to me toys! Aid your champion in crushing the humans "
            .. "who forever abuse you!\" crowed Princess."
        table.insert(combatstate.combatlog, msg)

        for i=1,numsummons do
            local summonname = i % 2 == 0 and "actionfigure" or "drone"
            local ally = scarlettables.copy(characters[summonname])
            ally.name = F"{ally.name}_{summoncount}_summon"
            for _, statistic in ipairs(ally.statistics._keys) do
                if statistic ~= "energy" and statistic ~= "maxenergy" then
                    local value = ally.statistics[statistic]
                    ally.statistics[statistic] = value + buff
                end
            end
            summoncount = summoncount + 1
            combatstateops.addenemy(combatstate, ally)
            table.insert(
                combatstate.combatlog,
                F"A {ally.combatname} joins the fray!"
            )
        end

        powerutils.inflictcost(combatstate, M.princess.summontoys)
    end
)

M.princess.summontoys.cost = 20
M.princess.summontoys.numtargets = 0
M.princess.summontoys.stats = {"power"}
M.princess.summontoys.damage = powers.NEITHER
M.princess.summontoys.description = "Princess summons a horde of toys to"
    .. " attack her opponents."
M.princess.summontoys.aiexecution = function() return M.princess.summontoys end
M.princess.summontoys.damage = powers.NEITHER
M.princess.summontoys.targets = combatstateops.activeallies

--[[
-- Grants the specified character the strongerfaster status, increasing attack
-- and defense.
--
-- user - The character granting the status
-- target - The character being buffed
-- combatstate - The current state of combat
--]]
local function bufftoy(user, target, combatstate)
    local userpower = attacks.statistic(user, "power")
    local statbuff = tostring(math.max(1,math.floor(userpower / 10)))
    local duration = attacks.statistic(user, "power")
    local description = target.combatname 
        .. " is moving faster, and swinging harder!"
    statusops.addstatus(target, statuses.strongerfaster, duration, user, combatstate)
end

--[[
-- Give all of Princess' toy minions a bonus to their strength and speed equal to
-- power/5 for power turns.
--]]
M.princess.strengthentoys=pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local activecharacter = combatstateops.activecharacter(combatstate)
        local name = activecharacter.combatname
        table.insert(
            combatstate.combatlog,
             name .. "' eyes flared and a wave of crimson light erupted "
                .. 'from her body. "Let my power FLOW THROUGH YOU!"'
        )
        local princessallies = combatstateops.activeallies(combatstate)
        for _, ally in ipairs(princessallies) do
            if ally.type == "toy" then
                bufftoy(activecharacter, ally, combatstate)
            end
        end
        powerutils.inflictcost(combatstate, M.princess.strengthentoys)
    end
)

M.princess.strengthentoys.cost = 4
M.princess.strengthentoys.numtargets = powers.ALL_ALLIES
M.princess.strengthentoys.damage = powers.NEITHER
M.princess.strengthentoys.stats = {"power"}
function M.princess.strengthentoys.description(character)
    local buff = math.floor(attacks.statistic(character, "power")/10)
    return character.combatname .. " increases the attack and defense of all "
    .. "allied toys by *" .. buff .. "*."
end
M.princess.strengthentoys.aiexecution = function() return M.princess.strengthentoys end
M.princess.strengthentoys.damage = powers.NEITHER
M.princess.strengthentoys.targets = combatstateops.activeallies

local function findmrcuddles(combatstate)
    for _, character in pairs(combatstateops.activeallies(combatstate)) do
        if character.combatname == "Mr. Cuddles" then
            return character
        end
    end
end

local function bearswipemessage(combatstate)
    local mrcuddles = findmrcuddles(combatstate)
    return mrcuddles
        and '"Crush them! Smite them! OBLITERATE THEM!" said Princess.'
        or ('"Get up Mr. Cuddles! Get up!" wailed Princess. '
            .. '"Smite my enemies, please!"')
end

--[[
-- Princess orders her bear to attack all allies. Doesn't do anything if
-- the bear is incapacitated.
--]]
M.princess.bearswipe=pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local mrcuddles = findmrcuddles(combatstate)
        table.insert(combatstate.combatlog, bearswipemessage(combatstate))
        if not mrcuddles then
            return
        end
        local opponents = combatstateops.activeopponents(combatstate)
        attacks.damage(combatstate, nil, opponents, mrcuddles)
        powerutils.inflictcost(combatstate, M.princess.bearswipe)
    end
)

M.princess.bearswipe.cost = 5
M.princess.bearswipe.numtargets = powers.ALL_OPPONENTS
M.princess.bearswipe.stats = {"power"}
M.princess.bearswipe.damage = powers.DAMAGES
M.princess.bearswipe.description = "Princess orders Mr. Cuddles to"
    .. " attack all her opponents at once."
M.princess.bearswipe.aiexecution = function() return M.princess.bearswipe end
M.princess.bearswipe.damage = powers.DAMAGES
M.princess.bearswipe.targets = combatstateops.activeopponents

local function inflictbearguardstatus(user, target, combatstate)
    local duration = target.statistics.power
    statusops.addstatus(target, statuses.bearguard, duration, user, combatstate)
end

--[[
-- Princess orders her bear to protect her from all attacks for
-- a number of turns equal to her power stat.
--]]
M.princess.bearshield=pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local activecharacter = combatstateops.activecharacter(combatstate)
        local mrcuddles = findmrcuddles(combatstate)
        if mrcuddles then
            local msg = "\"Eek! Save me Mr. Cuddles!\" shrieks Princess."
            table.insert(combatstate.combatlog, msg)
            inflictbearguardstatus(mrcuddles, activecharacter, combatstate)
        else
            local msg = "\"Poor Mr. Cuddles,\" says Princess quietly. "
                .. "\"Your sacrifice will not be in vain!\""
            table.insert(combatstate.combatlog, msg)
        end
        powerutils.inflictcost(combatstate, M.princess.bearshield)
    end
)

M.princess.bearshield.cost = 3
M.princess.bearshield.numtargets = 0
M.princess.bearshield.damage = powers.NEITHER
M.princess.bearshield.stats = {"power"}
M.princess.bearshield.description = "Mr. Cuddles wraps Princess in a shield equal to"
    .. " half his current energy, and sets Princess' defense to match his."
M.princess.bearshield.aiexecution = function() return M.princess.bearshield end
M.princess.bearshield.damage = powers.NEITHER
M.princess.bearshield.targets = powers.activeallies

--[[
-- Increases the target's energy by (2/3)user.statistics.power.
--
-- user: The healer
-- target: The healed
-- combatstate: The current state of combat
--]]
local function healmrcuddles(user, target, combatstate)
    local healing = math.floor(attacks.statistic(user, "power") * 2 / 3)
    local description = target.combatname .. " glows with newfound power!"
    table.insert(combatstate.combatlog, description)
    local maxenergy = target.statistics.energy
    local energy = target.statistics.energy
    target.statistics.energy = math.min(maxenergy, energy + healing) 
end

--[[
-- Transfers some of Princess' energy to Mr. Cuddles. The power
-- costs princess.statistics.power points, but Mr. Cuddles is
-- only healed for 2/3 of that (power transfer isn't 100% efficient).
--]]
M.princess.healbear=pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local user = combatstateops.activecharacter(combatstate)
        local target = findmrcuddles(combatstate)
        local msg = '"Take my power Mr. Cuddles! USE IT TO CRUSH OUR ENEMIES!"'
            .. " shrieks Princess."
        table.insert(combatstate.combatlog, msg)
        healmrcuddles(user, target, combatstate)
        powerutils.inflictcost(combatstate, M.princess.healbear)
    end
)

function M.princess.healbear.cost(character)
    return attacks.statistic(character, "power")
end 
M.princess.healbear.numtargets = 1
M.princess.healbear.stats = {"power"}
M.princess.healbear.damage = powers.HEALS
function M.princess.healbear.description(character)
    local healing = math.floor(2 * attacks.statistic(character, "power") / 3)
    return F"Princess heals Mr. Cuddles for *{healing}* energy."
end
M.princess.healbear.aiexecution = function() return M.princess.healbear end
M.princess.healbear.damage = powers.HEALS
M.princess.healbear.targets = combatstateops.activeallies

-------------- Mr Cuddles -------------
--[[
-- Mr. Cuddles is Princess' primary enforcer. He protects her,
-- attacks and disciplines her enemies.
--]]
M.mrcuddles = {}

--[[
-- Mr. Cuddles receives 2 * attack retaliation points.
--]]
M.mrcuddles.bearcounter = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local msg = F"{character.combatname} braces for attacks."
        table.insert(combatstate.combatlog, msg)
        local duration = attacks.statistic(character, "speed")
        statusops.addstatus(character, statuses.counter, duration, combatstate)
        powerutils.inflictcost(combatstate, M.mrcuddles.bearcounter)
    end
)

M.mrcuddles.bearcounter.cost = 3
M.mrcuddles.bearcounter.numtargets = 0
M.mrcuddles.bearcounter.stats = {"strength"}
M.mrcuddles.bearcounter.damage = powers.NEITHER
function M.mrcuddles.bearcounter.description(character)
    return "Mr. Cuddles gains retaliation points equal to "
    .. "*" .. 2 * attacks.attackstat(character) .. "* for "
    .. "*" .. attacks.statistic(character, "speed") .. "*"
    .. " turns."
end
M.mrcuddles.bearcounter.aiexecution = function() return M.mrcuddles.bearcounter end
M.mrcuddles.bearcounter.damage = powers.NEITHER
M.mrcuddles.bearcounter.targets = combatstateops.activeallies

--[[
-- Bearpunch allows the user to attack as if they had double the
-- attack modifier.
--
-- primitive: The damage primitive being executed
-- combatstate: The current state of combat
--
-- return The damage bearpunch should do, and a description of the punch
--]]
local function bearpunchdamage(user, target, combatstate)
    local function formula(attacker, defender)
        local attack = math.floor(1.5 * attacks.attackstat(attacker))
        local defense = attacks.defensestat(defender)
        return math.max(0, attack - defense)
    end
    local function damage(attacker, defender, combatstate)
        return attacks.defaultdamage(attacker, defender, combatstate, formula)
    end
    attacks.damage(combatstate, damage, {target}, user)
end

--[[
-- Given a list of opponents, returns the target to
-- target with bearpunch.
--
-- opponents: The opponents to select a target from
-- combatstate: The current state of combat
--
-- returns The target to punch
--]]
local function bearpunchtarget(opponents, combatstate)
    local _, target = next(opponents)
    if combatstate.round % 2 == 0 then
        for _, character in pairs(opponents) do
            if character.statistics.energy > target.statistics.energy then
                target = character
            end
        end
    else
        for _, character in pairs(opponents) do
            if character.statistics.energy < target.statistics.energy then
                target = character
            end
        end
    end
    return target
end

-- Bearpunch is a straightforward attack that does double Mr. Cuddle's attack
-- in damage, and inflicts Away on the target. It alternates between targeting the 
-- highest and lowest energy targets.
M.mrcuddles.bearpunch = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local msg = F"{character.combatname} unleashes a brutal punch!"
        table.insert(combatstate.combatlog, msg)
        local opponents = combatstateops.opponents(combatstate, character)
        local target = bearpunchtarget(opponents, combatstate)
        bearpunchdamage(character, target, combatstate)
        statusops.addstatus(target, statuses.away, 1, character, combatstate)
        powerutils.inflictcost(combatstate, M.mrcuddles.bearpunch)
    end
)

M.mrcuddles.bearpunch.cost = 5
M.mrcuddles.bearpunch.numtargets = 1
M.mrcuddles.bearpunch.stats = {"strength"}
M.mrcuddles.bearpunch.damage = powers.DAMAGES
function M.mrcuddles.bearpunch.description(character)
    return character.combatname
    .. " does up to "
    .. "*" .. math.floor(1.5 * attacks.attackstat(character)) .. "*"
    .. " damage to a target, and inflicts Away on them."
    .. " Damage is reduced by the target's speed."
end
M.mrcuddles.bearpunch.aiexecution = function() return M.mrcuddles.bearpunch end
M.mrcuddles.bearpunch.damage = powers.DAMAGES
M.mrcuddles.bearpunch.targets = combatstateops.activeopponents

-- [[
-- Mr. Cuddles grabs the character with the highest speed.
-- ]]
local function highestspeedtarget(opponents)
    local _, fastest = next(opponents)
    for _, opponent in ipairs(opponents) do
        local speed = attacks.statistic(opponent, "speed")
        if speed > attacks.statistic(fastest, "speed") then
            fastest = opponent
        end
    end
    return fastest
end

---------------- Beargrab ----------------
-- Mr. Cuddles grabs an opponent, dropping their
-- speed by Mr. Cuddles' strength/5. A grabbed character also recovers
-- automatically from Away.
M.mrcuddles.beargrab = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local opponents = combatstateops.opponents(combatstate, character)
        local target = highestspeedtarget(opponents)
        local msg = F"{character.combatname} grabs {target.combatname}!"
        table.insert(combatstate.combatlog, msg)
        local characterstrength = attacks.statistic(character, "strength")
        local targetstrength = attacks.statistic(target, "strength")
        local duration = math.max(1, characterstrength - targetstrength)
        statusops.addstatus(target, statuses.grabbed, duration, character, combatstate)
        powerutils.inflictcost(combatstate, M.mrcuddles.beargrab)
    end
)

M.mrcuddles.beargrab.cost = 10
M.mrcuddles.beargrab.numtargets = 1
M.mrcuddles.beargrab.stats = {"strength"}
M.mrcuddles.beargrab.damage = powers.NEITHER
function M.mrcuddles.beargrab.description(character)
    return character.combatname
        .. " grabs an opponent for up to "
        .. "*" .. attacks.statistic(character, "strength") .. "*"
        .. " turns. Upon being grabbed, a character's speed is reduced to "
        .. "*" .. math.floor(attacks.statistic(character, "strength") / 5) 
        .. "*. While grabbed, targets automatically recover from Away."
end
M.mrcuddles.beargrab.aiexecution = function() return M.mrcuddles.beargrab end
M.mrcuddles.beargrab.targets = combatstateops.activeopponents
M.mrcuddles.beargrab.damage = powers.NEITHER

--------- Bearspank
-- If a target is Grabbed, Mr. Cuddles can give them a good spanking, 
-- inflicting Spanked, and
-- reducing the duration of one buff.

local function findgrabbedtarget(combatstate)
    for _, opponent in pairs(combatstateops.activeopponents(combatstate)) do
        if statusops.statusduration(opponent, statuses.grabbed) then
            return opponent
        end
    end
end

M.mrcuddles.bearspank = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local user = combatstateops.activecharacter(combatstate)
        local target = findgrabbedtarget(combatstate)
        if not target then
            return M.mrcuddles.beargrab.transformer(combatstate)
        end
        local attackstrength = attacks.statistic(user, "strength")
        local targetstrength = attacks.statistic(target, "strength")
        local duration = math.max(1, attackstrength - targetstrength)
        local descriptions = {
            string.format(
                "%s rotates %s so that %s bottom is pointed at %s. %s's large"
                .. " paw repeatedly splats against %s's butt. %s kicks and squeals,"
                .. "  %s bottom flattening beneath %s's large hand.",
                user.combatname,
                target.combatname,
                powerutils.hisher(target),
                user.combatname,
                user.combatname,
                target.combatname,
                target.combatname,
                powerutils.hisher(target),
                user.combatname
            ),
            string.format(
                "%s holds %s upside down and strikes %s vulnerable  bottom"
                .. " with bum bobbling blows. %s wiggles and squirms, yelping"
                .. " with each stroke.",
                user.combatname,
                target.combatname,
                powerutils.hisher(target),
                target.combatname
            ),
            string.format(
                "%s tucks %s under %s arm and thrashes %s's outthrust bottom."
                .." %s kicks and wails. %s rapidly reddening bottom wobbles"
                .. " with each stroke.",
                user.combatname,
                target.combatname,
                powerutils.hisher(user),
                target.combatname,
                target.combatname,
                powerutils.HisHer(target)
            )
        }
        local msg = descriptions[math.random(#descriptions)]
        table.insert(combatstate.combatlog, msg)
        statusops.addstatus(target, statuses.spanked, duration, user, combatstate)
        powerutils.inflictcost(combatstate, M.mrcuddles.bearspank)
    end
)

M.mrcuddles.bearspank.cost = 5
M.mrcuddles.bearspank.numtargets = 1
M.mrcuddles.bearspank.stats = {"strength"}
M.mrcuddles.bearspank.damage = powers.NEITHER
M.mrcuddles.bearspank.description =  "Mr. Cuddles spanks a Grabbed target,"
    .. " inflicting Spanked and removing the target's best buff."
M.mrcuddles.bearspank.aiexecution = function() return M.mrcuddles.bearspank end
M.mrcuddles.bearspank.damage = powers.NEITHER
M.mrcuddles.bearspank.targets = combatstateops.activeopponents

M.stuffins = {}

------------ Noodle Whack ----------
--[[
-- A straight up damaging power that targets either a spanked
-- character, or the lowest energy character. Does double damage
-- against spanked targets.
--]]

-- [[
-- Given a list of characters, computes a target for noodle whack
--
-- opponents: The characters to pick a target from
--
-- returns the character to target with noodle whack
-- ]]
local function noodlewhacktarget(opponents)
    local _, target = next(opponents)
    for _, character in ipairs(opponents) do
        if statusops.statusduration(character, statuses.spanked) then
            return character
        end
        local energy = character.statistics.energy
        if energy < target.statistics.energy then
            target = character
        end
    end
    return target
end

local function noodlewhackdamage(user, target, combatstate)
    local function formula(attacker, defender)
        local attack = attacks.attackstat(attacker)
        if statusops.statusduration(target, statuses.spanked) then
            attack = math.floor(1.5 * attack)
        end
        local defense = attacks.defensestat(defender)
        return math.max(0, attack - defense)
    end
    local function damage(attacker, defender, combatstate1)
        return attacks.defaultdamage(attacker, target, combatstate1, formula)
    end
    attacks.damage(combatstate, damage, {target}, user)
end

M.stuffins.noodle_whack = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local opponents = combatstateops.activeopponents(combatstate)
        local target = noodlewhacktarget(opponents)
        local user = combatstateops.activecharacter(combatstate)
        local msg = F"{user.combatname} whacks {target.combatname} "
            .. "with a pool noodle!"
        table.insert(combatstate.combatlog, msg)
        noodlewhackdamage(user, target, combatstate)
        powerutils.inflictcost(combatstate, M.stuffins.noodle_whack)
    end
)

M.stuffins.noodle_whack.cost = 3
M.stuffins.noodle_whack.numtargets = 1
M.stuffins.noodle_whack.stats = {"strength"}
M.stuffins.noodle_whack.damage = powers.DAMAGES
M.stuffins.noodle_whack.description =  "Sir Stuffins whacks a target with"
    .. " his pool noodle. Does double damage to Spanked targets."
M.stuffins.noodle_whack.aiexecution = function() return M.stuffins.noodle_whack end
M.stuffins.noodle_whack.damage = powers.DAMAGES
M.stuffins.noodle_whack.targets = combatstateops.activeopponents

----------- Noodle Spank
-- A straightforward spanking. It inflicts Spanked and reduces the
-- target's best buff by sirstuffins.strength - target.strength.

--[[
-- Given a list of opponents, returns a target for Sir Stuffins to spank.
--
--opponents: The opponents to choose from
--
--return A target for Sir Stuffins to spank with his pool noodle.
--]]
local function noodlespanktarget(opponents)
    local _, loweststrengthtarget = next(opponents)
    for _, opponent in ipairs(opponents) do
        local strength = attacks.statistic(opponent, "strength")
        if strength < attacks.statistic(loweststrengthtarget, "strength") then
            loweststrengthtarget = opponent
        end
    end
    return loweststrengthtarget
end

local function noodlespankdescriptions(user, target)
    return {
        string.format(
            "%s flings %s over Giraffalot's neck and starts whaling on "
            .. "%s bottom with %s pool noddle. %s yelps and bucks before "
            .. "falling off.",
            user.combatname,
            target.combatname,
            powerutils.hisher(target),
            powerutils.hisher(user),
            target.combatname
        ),
        string.format(
            "Giraffalot pins %s facedown "
            .. "against the ground. %s leans over and smacks %s a dozen "
            .. "times right on %s vulnerable tushy. %s yelps and kicks. ",
            target.combatname,
            user.combatname,
            target.combatname,
            powerutils.hisher(target),
            target.combatname
        )
    }
end

M.stuffins.noodle_spank = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local opponents = combatstateops.activeopponents(combatstate)
        local target = noodlespanktarget(opponents)
        local user = combatstateops.activecharacter(combatstate)
        local userstrength = attacks.statistic(user, "strength")
        local targetstrength = attacks.statistic(target, "strength")
        local duration = math.max(1, userstrength - targetstrength)
        local descriptions = noodlespankdescriptions(user, target)
        local description = descriptions[math.random(#descriptions)]
        table.insert(combatstate.combatlog, description)
        statusops.addstatus(target, statuses.spanked, duration, user, combatstate)
        powerutils.inflictcost(combatstate, M.stuffins.noodle_spank)
    end
)

M.stuffins.noodle_spank.cost = 5
M.stuffins.noodle_spank.numtargets = 1
M.stuffins.noodle_spank.stats = {"strength"}
M.stuffins.noodle_spank.damage = powers.NEITHER
M.stuffins.noodle_spank.description =  "Sir Stuffins spanks a target with"
    .. " his pool noodle. Inflicts Spanked, and reduces duration of"
    .. " the target's best buff."
M.stuffins.noodle_spank.aiexecution = function() return M.stuffins.noodle_spank end
M.stuffins.noodle_spank.damage = powers.NEITHER
M.stuffins.noodle_spank.targets = combatstateops.activeopponents

-------------- Charge ----------------
--[[
-- Mr. Stuffins charges an opponent while riding Giraffalot.
-- Can't be used if Mr. Stuffins or Giraffalot are Away (because
-- then Mr. Stuffins isn't on Giraffalot!)
--
-- When charging, Mr. Stuffins performs an attack that adds
-- Giraffalot's speed to the damage, and inflicts Away on the target.
-- Charge targets the enemy with the lowest defense (i.e. the one who
-- he can do the most damage to).
--]]

local function chargeformula(user, target, giraffalot)
    local giraffalotspeed = attacks.statistic(giraffalot, "speed")
    local attack = attacks.attackstat(user) + giraffalotspeed
    local defense = attacks.defensestat(target)
    return math.max(0, attack - defense)
end

local function chargetarget(targets)
    local besttarget = next(targets)
    local worstdefense = attacks.defensestat(besttarget)
    for _, target in ipairs(targets) do
        local defense = attacks.defensestat(target)
        if defense < worstdefense then
            besttarget = target
            worstdefense = defense
        end
    end
    return besttarget
end

local function chargedamage(attacker, defender, combatstate)
    local giraffalot = combatutils.findcharacterbycombatname(
        combatstate,
        require('data/characters').stuffed_giraffe.combatname
    )
    local function formula(user, target)
        return chargeformula(user, target, giraffalot)
    end
    return attacks.defaultdamage(attacker, defender, combatstate, formula)
end

M.stuffins.charge = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local user = combatstateops.activecharacter(combatstate)
        local giraffalot = combatstateops.findcharacterbycombatname(
            combatstate,
            require('data/characters').stuffed_giraffe.combatname
        )
        local isaway = statusops.statusduration(user, statuses.away)
            or statusops.statusduration(giraffalot, statuses.away)
        local giraffalotdefeated = combatstateops.isdefeated(
            combatstate,
            giraffalot.name
        )
        if isaway or giraffalotdefeated then
            table.insert(
                combatstate.combatlog,
                string.format(
                    "%s has to be mountd on %s to charge!",
                    user.combatname,
                    giraffalot.combatname
                )
            )
            return
        end
        local target = chargetarget(opponents)
        chargedamage(user, target, combatstate)
        statusops.addstatus(target, statuses.away, 1, user, combatstate)
        powerutils.inflictcost(combatstate, M.stuffins.charge)
    end
)

M.stuffins.charge.cost = 10
M.stuffins.charge.numtargets = 1
M.stuffins.charge.stats = {"strength"}
M.stuffins.charge.damage = powers.DAMAGES
M.stuffins.charge.description =  "Sir Stuffins charges a target."
    .. " Does damage based on his attack and Giraffalot's speed."
    .. " Inflicts Away. Does nothing if he or Giraffalot are Away,"
    .. " or Giraffalot is defeated."
M.stuffins.charge.aiexecution = function() return M.stuffins.charge end
M.stuffins.charge.targets = combatstateops.activeopponents
M.stuffins.charge.damage = powers.DAMAGES

M.giraffalot = {}

------------- Rear
--[[
-- Giraffalot rears, doubling her speed for two rounds.
--]]
M.giraffalot.rear = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        statusops.addstatus(character, statuses.reared, 2, character, combatstate)
        table.insert(
            combatstate.combatlog, 
            F"{character.combatname} rears up on two legs!"
        )
        powerutils.inflictcost(combatstate, M.giraffalot.rear)
    end
)

M.giraffalot.rear.cost = 5
M.giraffalot.rear.numtargets = 0
M.giraffalot.rear.stats = {"speed"}
M.giraffalot.rear.damage = powers.NEITHER
M.giraffalot.rear.description =  "Giraffalot rears, preparing to charge."
    .. " Doubles Giraffalot's speed for 2 rounds."
M.giraffalot.rear.aiexecution = function() return M.giraffalot.rear end
M.giraffalot.rear.targets = combatstateops.activeallies
M.giraffalot.rear.damage = powers.NEITHER

-------------- Trample ---------------
--[[
-- Giraffalot tramples an opponent. Giraffalot's speed
-- is added to her attack for the power. A trampled
-- opponent has their speed reduced by
-- giraffalot.statistics.speed - (target.statistics.speed / 2)
-- for 2 turns.
--
-- Giraffalot targets the character with the lowest non-zero speed for
-- this attack.
--]]

local function trampletarget(opponents)
    local _, lowestnonzero = next(opponents)
    for _, character in ipairs(opponents) do
        local speed = attacks.statistic(character, "speed")
        if speed > 0 and speed < attacks.statistic(lowestnonzero, "speed") then
            lowestnonzero = character
        end
    end
    return lowestnonzero
end

local function trampleformula(user, target)
    local speed = attacks.statistic(user, "speed")
    local attack = attacks.attackstat(user) + speed
    local defense = attacks.defensestat(target)
    return math.max(0, attack - defense)
end

local function trampledamage(user, target, combatstate)
    return attacks.defaultdamage(user, target, combatstate, trampleformula)
end

M.giraffalot.trample = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local user = combatstateops.activecharacter(combatstate)
        local opponents = combatstateops.activeopponents(combatstate)
        local target = trampletarget(opponents)
        table.insert(
            combatstate.combatlog,
            user.combatname .. " tramples " .. target.combatname .. "!"
        )
        attacks.damage(combatstate, trampledamage, {target}, user)
        local userspeed = attacks.statistic(user, "speed")
        local targetspeed = attacks.statistic(target, "speed")
        local debuff = userspeed - math.floor(targetspeed / 2)
        local slow = statuses.speed(debuff)
        statusops.addstatus(target, statuses.fast(-debuff), 2, user, combatstate)
        powerutils.inflictcost(combatstate, M.giraffalot.trample)
    end
)

M.giraffalot.trample.cost=7
M.giraffalot.trample.numtargets=1
M.giraffalot.trample.stats={"strength", "speed"}
M.giraffalot.trample.damage=powers.DAMAGES
function M.giraffalot.trample.description(character)
    local speed = attacks.statistic(character, "speed")
    local attack = attacks.attackstat(character) + speed
    return "Giraffalot tramples an opponent, doing up to *" .. attack .. "*"
        .. " damage to them, and reducing their speed for 2 turns."
end
M.giraffalot.trample.aiexecution = function() return M.giraffalot.trample end
M.giraffalot.trample.targets = combatstateops.activeopponents
M.giraffalot.trample.damage = powers.DAMAGES

-- [[
-- Returns the opponent most vulnerable to pin.
--
-- The opponent most vulnerable to pin is the one with the
-- lowest strength.
--
-- opponents: The opponents to select a target from
--
-- return The character in opponents with the lowest strength
-- ]]
local function pintarget(opponents)
    local _, besttarget = next(opponents)
    local besttargetstrength = attacks.statistic(besttarget, "strength")
    for _, target in ipairs(opponents) do   
        targetstrength = attacks.statistic(target, "strength")
        if targetstrength < besttargetstrength then
            besttarget = target
        end
    end
    return besttarget
end

-------------------- Pin --------------
--[[
-- Giraffalot jumps on an opponent, doing damage
-- based on her strength and inflicting Grabbed on the
-- target for a duration based on relative strength.
--]]
M.giraffalot.pin = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local user = combatstateops.activecharacter(combatstate)
        local opponents = combatstateops.activeopponents(combatstate)
        local target = pintarget(opponents)
        table.insert(
            combatstate.combatlog,
            user.combatname .. " jumps on " .. target.combatname .. "!"
        )
        attacks.damage(combatstate, nil, {target}, user)
        local userstrength = attacks.statistic(user, "strength")
        local targetstrength = attacks.statistic(target, "strength")
        local duration = math.max(1, userstrength - targetstrength)
        statusops.addstatus(target, statuses.grabbed, duration, user, combatstate)
        powerutils.inflictcost(combatstate, M.giraffalot.pin)
    end
)

M.giraffalot.pin.cost=3
M.giraffalot.pin.numtargets=1
M.giraffalot.pin.stats={"strength"}
M.giraffalot.pin.damage=powers.DAMAGES
function M.giraffalot.pin.description(character)
    local attack = attacks.attackstat(character)
    return "Giraffalot jumps on an opponent and pins them,"
        .. " doing damage based on his attack, and inflicting Grabbed"
        .. " for a duration based on strength."
end
M.giraffalot.pin.aiexecution = function() return M.giraffalot.pin end
M.giraffalot.pin.damage = powers.DAMAGES
M.giraffalot.pin.targets = combatstateops.activeopponents

M.legostopheles = {}

--- Grab
--- Inflicts Grabbed on the target. While the target is grabbed,
-- Legostopheles' attack count is reduced by 1. Targets the
-- enemy with the highest speed.
M.legostopheles.grab = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local user = combatstateops.activecharacter(combatstate)
        local opponents = {}
        for _, enemy in ipairs(combatstateops.activeopponents(combatstate)) do
            if not statusops.statusduration(enemy, statuses.grabbed) then
                table.insert(opponents, enemy)
            end
        end
        local target = highestspeedtarget(opponents, combatstate)
        local userstrength = attacks.statistic(user, "strength")
        local targetstrength = attacks.statistic(target, "strength")
        local duration = math.max(2, userstrength - targetstrength)
        table.insert(
            combatstate.combatlog,
            user.combatname .. " grabs " .. target.combatname 
                .. " with a lego tentacle!"
        )
        statusops.addstatus(target, statuses.grabbed, duration, user, combatstate)
        statusops.addstatus(user, statuses.reduceattack, duration, user, combatstate)
        powerutils.inflictcost(combatstate, M.legostopheles.grab)
    end
)

M.legostopheles.grab.cost=3
M.legostopheles.grab.numtargets=1
M.legostopheles.grab.stats={"strength"}
function M.legostopheles.grab.description(character)
    return "Legostopheles inflicts Grabbed on a target"
    .. " for a duration based on relative strength."
    .. " When Grabbed, a character's speed is reduced,"
    .. " and they are immune to Away."
    .. " Legostopheles' attack count is reduced by 1 while"
    .. " the target is Grabbed."
end
M.legostopheles.grab.aiexecution = function() return M.legostopheles.grab end
M.legostopheles.grab.damage=powers.NEITHER
M.legostopheles.grab.targets = combatstateops.activeopponents

------------- Lego Swirl ------------
--[[
-- Legostopheles sends a swarm of legos to swirl around the target.
-- The legos interfere with the target's attack, reducing it by
-- by legostopheles.strength / 5, for a duration based on relative strength.
-- It also reduces Legostopheles' attack count by 1. Targets the enemy
-- with the highest attack.
--]]

local function inflictlegoswirl(user, target, combatstate)
    local duration = attacks.statistic(user, "strength")
    table.insert(
        combatstate.combatlog,
        string.format(
            "One of Legostopheles' tentacles disintigrates into legos "
            .. "and swirl around %s, buffeting and blinding %s!",
            target.combatname,
            powerutils.himher(target)
        )
    )
    statusops.addstatus(target, statuses.legoswirl, duration, user, combatstate)
    statusops.addstatus(user, statuses.reduceattack, duration, user, combatstate)
end

local function higheststrengthtarget(opponents)
    local strength = -1
    local strongest
    for _, target in ipairs(opponents) do
        local targetstrength = attacks.statistic(target, "strength")
        if targetstrength > strength then
            strength = targetstrength
            strongest = target
        end
    end
    return strongest
end

M.legostopheles.legoswirl = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local user = combatstateops.activecharacter(combatstate)
        local opponents = {}
        for _, enemy in ipairs(combatstateops.activeopponents(combatstate)) do
            if not statusops.statusduration(enemy, statuses.legoswirl) then
                table.insert(opponents, enemy)
            end
        end
        local target = higheststrengthtarget(opponents)
        inflictlegoswirl(user, target, combatstate)
        powerutils.inflictcost(combatstate, M.legostopheles.legoswirl)
    end
)
M.legostopheles.legoswirl.cost=3
M.legostopheles.legoswirl.numtargets=1
M.legostopheles.legoswirl.stats={"strength"}
function M.legostopheles.legoswirl.description(character)
    local strength = attacks.statistic(character, "strength")
    return string.format(
        "Legostopheles sends a swarm of legos to swirl around a"
        .. " target. The legos distract and disorient the target,"
        .. " reducing their attack by *%d* for *%d*"
        .. " rounds. Reduces"
        .. " Legostopheles' Attack count by 1.",
        math.floor(strength / 5),
        strength
    )
end
M.legostopheles.legoswirl.aiexecution = function() return M.legostopheles.legoswirl end
M.legostopheles.legoswirl.damage = powers.NEITHER
M.legostopheles.legoswirl.targets = combatstateops.activeopponents

------------- Lego Paddle ------------

local function legopaddledescriptions(spanker, spankee)
    local spankername = spanker.combatname
    local spankeename = spankee.combatname
    return {
       F"{spankername} holds {spankeename} upside down and batters {powerutils.hisher(target)} round bottom with swift, stinging strokes!",
       F"{spankername} tucks {spankeename} under its tentacle and peppers {spankeename}'s vulnerable bottom with its giant lego paddle!" 
    }
end

-- Legostopheles swirls one of his tentacles into a paddle. Decreases his
-- attack count by 1, and increases his spank count by 1.
-- Duration is equal to Legostopheles' strength.
M.legostopheles.legopaddle = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local user = combatstateops.activecharacter(combatstate)
        local duration = attacks.statistic(user, "strength")
        statusops.addstatus(
            user, 
            statuses.increasespank(legopaddledescriptions), 
            duration, 
            user, 
            combatstate
        )
        statusops.addstatus(user, statuses.reduceattack, duration, user, combatstate)
        powerutils.inflictcost(combatstate, M.legostopheles.legopaddle)
    end
)

M.legostopheles.legopaddle.cost=3
M.legostopheles.legopaddle.numtargets=0
M.legostopheles.legopaddle.stats={"strength"}
M.legostopheles.legopaddle.damage=powers.NEITHER
M.legostopheles.legopaddle.description=
    "Legostopheles swirls"
    .. " one of his tentacles into a paddle, increasing"
    .. " his spank count by 1, and reducing his attack"
    .. " count by 1. He only paddles Grabbed targets."
M.legostopheles.legopaddle.aiexecution = function() return M.legostopheles.legopaddle end
M.legostopheles.legopaddle.damage = powers.NEITHER
M.legostopheles.legopaddle.targets = combatstateops.activeopponents

return M
