--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--
--[[
-- Defines a collection of helper methods for working with UI story pages.
-- These are defined in a separate module in order to make them easier to
-- test. None of these functions require the love or SUIT frameworks, nor
-- do they require any data files like font files. This is important, because
-- different OS's seem to handle relative files differently so that tests
-- on UI functions that work on one OS may not work on another OS.
--]]

local M = {}

-- The maximum number of characters allowed in each line of text.
M.MAX_LINE_LENGTH = 80

local log = require 'log'

local scarlettables = require 'scarlettables'

local F = require 'f-strings/F'

--[[
-- Given a string representing a paragraph, and the length of each line,
-- splits the paragraph into a list of strings such that each line in the
-- list is no longer than the line length.
--
-- paragraph: The text to wrap
-- linelength: The maximum number of characters allowed on each line
--
-- Returns a list of strings such that each string is a line in paragraph no
-- longer than linelength.
--]]
function M.wrapparagraph(paragraph, linelength)
    local wrapped = {}
    local line = {}
    local currentlength = 0
    for word in string.gmatch(paragraph, "%S+") do
        local wordlength = string.len(word)
        -- need to add 1 to account for a space after the word.
        currentlength = currentlength + wordlength + 1
        -- We don't count the space after the last word, because that won't
        -- be printed.
        if currentlength - 1 <= linelength then
            table.insert(line, word)
        else
            -- Double space in the LOVE UI, single space otherwise
            local textline = table.concat(line, " ") .. (LOVE and "\n" or "")
            table.insert(wrapped, textline)
            line = {word}
            currentlength = wordlength + 1
        end
    end
    table.insert(wrapped, table.concat(line, " ") .. "\n")
    return wrapped
end

--[[
-- Given a list of paragraphs to display, splits each paragraph into a list
-- of lines such that each line is short enough to comfortably display on the
-- game screen. Each line will be no longer than LINE_LENGTH, though they
-- may be shorter if the current screen is not wide enough to display
-- LINE_LENGTH characters.
--
-- paragraphs:  The list of strings to wrap
-- screenlinelength: The maximum number of characters that can comfortably
-- fit on the screen
--
-- Returns a list of lists of strings, where each entry is a paragraph, and each
-- paragraph is a list of lines short enough to comfortably display in a
-- single line on the gamescreen
--]]
function M.wraptext(paragraphs, screenlinelength, maxlength)
    maxlength = maxlength or M.MAX_LINE_LENGTH
    -- The number of characters we can comfortably show in one line
    log.trace("Screen line length: " .. screenlinelength)
    log.trace("max line length " .. maxlength)
    local wrappedparagraphs = {}
    for _, paragraph in ipairs(paragraphs) do
        table.insert(
            wrappedparagraphs,
            M.wrapparagraph(paragraph, math.min(screenlinelength, maxlength))
        )
    end
    return wrappedparagraphs
end

--[[
-- Given a list of paragraphs representing the current text to display
-- such that each paragraph is a list of lines, truncates the text down to only
-- those lines that are visible.
--
-- storytext: A list of lists of strings, each entry in the list is a paragraph
-- startline: The line of text that we are currently displaying at the top of
--           the screen, if -1 then it's whatever line is needed so that the
--           last line in the text is the last line on the screen
-- maxlinenumber: The maximum number of lines that can be displayed on the
--  screen
-- Returns a sublist of storytext such that the first entry is the line at
--  line number startline, and the last is at line number startline + the
--  maximum number of lines the screen can display
--]]
function M.truncate(storytext, startline, maxlinenumber)
    startline = startline or 1
    -- Scroll to the bottom if starttline is -1.
    if startline == -1 then
        local totallines = 0
        for _, paragraph in ipairs(storytext) do
            for _, line in ipairs(paragraph) do
                totallines = totallines + 1
            end
        end
        startline = math.max(1, totallines - maxlinenumber)
    end
    local displayedText = {}
    local textbefore = startline > 1
    local linecounter = 1
    -- We add lines from the story text to the truncated text until we've
    -- reached or exceeded the maximum number of lines that we can display.
    -- Once we've reached that limit, we return the truncated text.
    for i, paragraph in ipairs(storytext) do
        if linecounter + #paragraph >= startline then
            table.insert(displayedText, {})
        end
        for j, line in ipairs(paragraph) do
            if linecounter >= startline then
                local currentparagraph = displayedText[#displayedText]
                table.insert(currentparagraph, line)
            end
            linecounter = linecounter + 1
            if linecounter - startline >= maxlinenumber then
                local textafter = i < #storytext or j < #storytext[#storytext]
                return displayedText, textbefore, textafter
            end
        end
        -- There is a blank line after each paragraph.
        linecounter = linecounter + 1
        if linecounter - startline >= maxlinenumber then
            local textafter = i < #storytext
            return displayedText, textbefore, textafter
        end
    end
    return displayedText, textbefore, false
end

--[[
-- Given a UI state, builds the choices to
-- display to the user
--
-- uistate: The current state of the ui
--]]
function M.buildstorychoices(uistate)
    local fonts = require "ui/fonts"
    local gui = uistate.gui

    local choicetext = "(.) Select Choice"

    local commandtext = "(Spacebar)Commands"
    local maxlength = 0
    local buttontext = { choicetext, commandtext }
    local buttonfont = fonts.button(uistate.core.gamestate.options.buttonfont)
    for _, text in ipairs(buttontext) do
        local textlength = fonts.width(buttonfont, text)
        if textlength > maxlength then
            maxlength = textlength
        end
    end
    local buttonheight = buttonfont:getHeight()
    gui.Button(
        choicetext,
        {font=buttonfont, id="."},
        gui.layout:row(buttonfont:getWidth(choicetext) + 5, buttonfont:getHeight())
    )

    gui.Button(
        commandtext,
        {font=buttonfont, id="space"},
        gui.layout:col(fonts.width(buttonfont, commandtext), buttonheight)
    )
end

--[[
-- Given the current state of the UI, returns a combat page if the specified
-- story page is a combat page, and otherwise returns `nil`
--
-- storyname - The page describing combat
-- uistate - The current state of the game
--
-- returns A UI page that starts combat
--]]
function M.combat(uistate, storypage)
    local combat = require "combat/combatui"
    local gamestate = uistate.core.gamestate
    local stories = uistate.core.gamestate.configuration.stories
    return combat.startcombat(
        uistate.core.gamestate,
        stories[storypage.victory],
        stories[storypage.defeat],
        M.resolvecharacters(gamestate, storypage.allies),
        M.resolvecharacters(gamestate, storypage.enemies)
    )
end

function M.resolvecharacters(gamestate, characternames)
    local characters = {}
    for i, name in pairs(characternames) do
        local character = assert(
            gamestate.characters[name],
            "Unknown character :'" .. name .. "'"
        )
        characters[i] = scarlettables.copy(character)
    end
    return characters
end


-- [[
-- Returns the total number of lines in the current event.
--
-- uistate: The current state of the UI
--
-- return the total number of lines for the given event
-- ]]
function M.numlines(uistate)
    local love = require 'love'
    local fonts = require 'ui/fonts'
    local bodyfontsize = uistate.core.gamestate.options.bodyfont
    local charwidth = fonts.singleCharacterWidth(fonts.body(bodyfontsize))
    local screenwidth = love.graphics.getWidth()
    log.trace("Screen width: " .. screenwidth)
    log.trace("Character width: " .. charwidth)
    local screenlinelength = math.floor((screenwidth - charwidth) / charwidth)
    local paragraphs = M.wraptext(
        uistate.core.gamestate.gametext,
        screenlinelength
    )
    local numlines = 0
    for _,paragraph in ipairs(paragraphs) do
        numlines = numlines + #paragraph
    end
    -- We have an empty line after each paragraph.
    numlines = numlines + #paragraphs
    return numlines
end


--[[
-- Given a paragraph, returns true if the paragraph has just empty lines,
-- false otherwise.
--
-- paragraph - The paragraph to check
--
-- returns true if the paragraph is blank, false otherwise
--]]
local function isblank(paragraph)
    for _, line in pairs(paragraph) do
        if string.match(line, "%S") then
            return false
        end
    end
    return true
end

-- [[
-- Given a collection of paragraphs, strips off the first and last if they
-- only contain blank lines.
--
-- paragraphs: The list of paragraphs to trim
-- ]]
local function trim(paragraphs)
    if paragraphs[1] then
        if isblank(paragraphs[1]) then
            table.remove(paragraphs, 1)
        end
        if isblank(paragraphs[#paragraphs]) then
            table.remove(paragraphs)
        end
    end
end

-- [[
-- Returns the number of characters that can fit on one line of text.
--
-- fonts: A table containing utility functions for working with fonts. The
--  table must have the following functions:
--      1. singleCharacterWidth - Takes a font and returns the max number of
--          pixels a single character takes in that font
-- font: The font for which we want to know the number of characters that will
--  fit
-- screenwidth: The width of the screen in pixels
--
-- returns The number of characters that will fit on a line with this font
-- ]]
local function numcharsperline(fonts, font, screenwidth)
    local charwidth = fonts.singleCharacterWidth(font)
    local numchars = (screenwidth - charwidth) / charwidth
    if numchars > M.MAX_LINE_LENGTH then
        numchars = M.MAX_LINE_LENGTH
    end
    return numchars
end


--[[
-- Builds a label that displays a segment of multiple paragraphs of text.
--
-- This also includes some arrows at the top and bottom of the label
-- indicating that the user can scroll up and down with the arrow keys.
--
-- uistate - The current state of the UI. The UI must include the following
--  keys:
--      line - The current line in the text that we are beginning our display
--      numlines - The total number of lines of text to display
--
--  Note: This function updates `uistate.numlines` with the total number of
--      lines that can be displayed.
--
--  love - The love game engine
--  font - The font to use to display the text
--  text - A list of strings to display
--  choices- A list of strings representing the choices available to the player
--  screenratio - The amount of screen space the body should take up
--  maxlength - The maximum length allowed for each line
--  scrollToBottom - Whether to automatically scroll so that the last line of
--  text is the last line to display
-- ]]
function M.buildBodyLabel(
    uistate,
    love,
    fonts,
    text,
    choices,
    screenratio,
    maxlength
)
    maxlength = maxlength or M.MAX_LINE_LENGTH
    local options = uistate.core.gamestate.options
    local bodyfont = fonts.body(options.bodyfont)
    local charwidth = fonts.singleCharacterWidth(bodyfont)
    local screenwidth, screenheight = love.graphics.getDimensions()
    local numcharsline = numcharsperline(fonts, bodyfont, screenwidth)
    local labelheight = screenheight * screenratio
    local numlines = 0
    local wrappedtext = M.wraptext(text, numcharsline, maxlength)
    local wrappedchoices = M.wraptext(choices, numcharsline, maxlength)
    local flattenedchoices = {}
    for _, choice in ipairs(wrappedchoices) do
        for _, line in ipairs(choice) do
            table.insert(flattenedchoices, line)
        end
    end
    if next(flattenedchoices) then
        table.insert(wrappedtext, flattenedchoices)
    end

    for _, paragraph in ipairs(wrappedtext) do
        numlines = numlines + #paragraph
    end
    -- We have an empty line after each paragraph.
    numlines = numlines + #wrappedtext
    uistate.numlines = numlines
    log.trace("numlines: " ..uistate.numlines)

    trim(wrappedtext)
    uistate.maxlinesonscreen = labelheight / (2 * bodyfont:getHeight()) - 1
    local paragraphs, textbefore, textafter = M.truncate(
        wrappedtext,
        uistate.line,
        uistate.maxlinesonscreen
    )
    uistate.textbefore = textbefore
    uistate.textafter = textafter
    for i, paragraph in ipairs(paragraphs) do
        paragraphs[i] = table.concat(paragraph, "\n")
    end

    local pixellength = numcharsline * charwidth
    local x, y, w, h = uistate.gui.layout:row(pixellength, labelheight)
    uistate.gui.Label(
        table.concat(paragraphs, "\n\n\n"),
        {
            id="story",
            font=bodyfont,
            align="left",
            valign="top"
        },
        x, y, w, h
    )

    local invalidtext = string.rep("---", 3)
    local downtext = "\\/ "
    local buttonfont = fonts.button(options.buttonfont)
    local width = math.max(
        fonts.width(buttonfont, invalidtext),
        fonts.width(buttonfont, "PageDown")
    )
    local buttontext = textafter and downtext or invalidtext
    local bx, by, bw, bh = uistate.gui.layout:row(width, buttonfont:getHeight())
    uistate.gui.Button(
        buttontext,
        {id = "down", font = buttonfont},
        bx, by, bw, bh
    )

    uistate.gui.layout:push(bx + width + uistate.gui.layout:padding(), by)
    buttontext = textafter and "PageDown" or invalidtext
    uistate.gui.Button(
        buttontext,
        {id = "pagedown", font = buttonfont},
        uistate.gui.layout:col(width, buttonfont:getHeight())
    )
    buttontext = "End"
    uistate.gui.Button(
        buttontext,
        {id = "end", font = buttonfont},
        uistate.gui.layout:col(width, buttonfont:getHeight())
    )
    buttontext = textbefore and "/\\" or invalidtext
    uistate.gui.Button(
        buttontext,
        {id = "up", font = buttonfont},
        uistate.gui.layout:col(width, buttonfont:getHeight())
    )
    buttontext = textbefore and "PageUp" or invalidtext
    uistate.gui.Button(
        buttontext,
        {id = "pageup", font = buttonfont},
        uistate.gui.layout:col(width, buttonfont:getHeight())
    )
    buttontext = "Home"
    uistate.gui.Button(
        buttontext,
        {id = "home", font = buttonfont},
        uistate.gui.layout:col(width, buttonfont:getHeight())
    )
    uistate.gui.layout:pop()
end

local function buildStatisticsPanelWidgit(fonts, uistate, font, lines)
    local text = table.concat(lines, "\n")
    uistate.gui.Label(
        text,
        {
            id="statistics",
            font=font,
            align="left",
            valign="top"
        },
        uistate.gui.layout:row(
            fonts.width(font, text),
            font:getHeight() * #lines
        )
    )
end

local bumemojis = {
[0]="(^_^)",
"(-_-)",
"(o_o)",
"(>_<)",
"(x_x)",
"(;_;)",
"(T_T)",
}

--[[
-- Given a list of strings containing bum status,
-- returns the text to display. For now, this is
-- an emoji. It could also be the list of entries.
--
-- bumstatus: A list of strings describing the current
--  state of the player's bottom
-- emojis: A boolean that should be true if the player wants to
-- display bum status as emojis, false otherwise
--
--  return A string to display to the player
--]]
function M.buildBumText(bumstatus, emojis)
    if emojis then
        local numstatuses = math.min(bumstatus and #bumstatus or 0, #bumemojis)
        return bumemojis[numstatuses]
    end
    return bumstatus
        and table.concat(bumstatus, "; ")
        or "Your bottom is smooth and unmarked. Hope it stays that way!"
end

function M.buildStatisticsPanel(fonts, love, uistate)
    local gamestate = uistate.core.gamestate
    local font = fonts.combatboard(gamestate.options.combatfont)
    local pc = gamestate.characters.pc
    local statistics = pc.statistics
    local lines = {}
    for _, statistic in ipairs(statistics._keys) do
        if statistic == "maxenergy" then
            lines[#lines] = lines[#lines] .. "/" .. statistics[statistic]
        else
            table.insert(lines, statistic .. ": " .. statistics[statistic])
        end
    end
    local labelfont = fonts.combatboard(gamestate.options.combatfont + 4)
    uistate.gui.Label(
        "Statistics",
        {id="statisticslabel", font=labelfont, align="center", valign="top"},
        uistate.gui.layout:row(
            fonts.width(labelfont, "Statistics"), labelfont:getHeight()
        )
    )
    buildStatisticsPanelWidgit(fonts, uistate, font, lines)
    local clothing = pc.clothing
    lines = {"(F1) hairstyle:" .. pc.attributes.hairstyle}
    local count = 2
    for _, slot in ipairs(clothing._keys) do
        local key = string.format("(F%d) ", count)
        count = count + 1
        assert(
            clothing[slot], 
            "Clothing not found in slot: " .. slot .. " slots: " .. scarlettables.string(clothing)
        )
        table.insert(lines, key .. slot .. ": " .. clothing[slot].name)
    end
    uistate.gui.Label(
        "Appearance",
        {id="appearancelabel", font=labelfont, align="center", valign="top"},
        uistate.gui.layout:row(
            fonts.width(labelfont, "Appearance"), labelfont:getHeight()
        )
    )
    buildStatisticsPanelWidgit(fonts, uistate, font, lines)
    local emojis = uistate.core.gamestate.options.bumstatus == "emoji"
    local bumtext = M.buildBumText(pc.bummarks, emojis)
    local availablewidth = love.graphics.getDimensions() / 3
    local numcharsline = numcharsperline(
        fonts,
        font,
        availablewidth
    )
    local wrappedbumtext = table.concat(
        M.wrapparagraph(bumtext, numcharsline),
        "\n"
    )
    uistate.gui.Label(
        wrappedbumtext,
        {id="bumtext", font=font, align="left", valign="top"},
        uistate.gui.layout:row(
            fonts.width(font, wrappedbumtext), font:getHeight()
        )
    )
end

return M
