--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--
-- [[
-- This module contains the dynamic text functions that will be used in this
-- game's story.
--]]
local dynamictext = require "story/dynamictext"
local log = require "log"
local scarlettables = require 'scarlettables'

local M = {}

local function getcharacter(gamestate, name)
    return assert(
        gamestate.characters[name],
        "Unknown character: " .. '"' .. (name or "nil") .. '"'
    )
end

function M.attribute(attributename, value, textifequal, textotherwise, name)
    name = name or "pc"
    textifequal = textifequal or ""
    textotherwise = textotherwise or ""
    log.trace(
        "Building attribute-based dynamic text: ",
        attributename, ";",
        value, ";",
        textifequal, ";",
        textotherwise, ";",
        name
    )
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        return dynamictext.attribute(
            character,
            attributename,
            value,
            textifequal,
            textotherwise
        )
    end
end

--[[
-- Given the name of a character, returns a function that chooses between two strings depending
-- on whether the specified character is a male or not.
--
-- textifmale: The text to return if the character is male
-- textotherwise: The text to return otherwise
-- name: The name of the character whose gender is of interest, defaults to "pc"
--]]
function M.gender(textifmale, textotherwise, name)
    name = name or "pc"
    log.trace(
        "Building gender-based dynamic text: ",
        textifmale, ";",
        textotherwise, ";",
        name
    )
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        return dynamictext.attribute(
            character,
            "gender",
            "male",
            textifmale,
            textotherwise
        )
    end
end

function M.ismale(textifmale, textotherwise, name)
    return M.gender(textifmale, textotherwise, name)
end

function M.male(textifmale, textotherwise, name)
    return M.ismale(textifmale, textotherwise, name)
end

function M.isfemale(textiffemale, textifmale, name)
    return M.gender(textifmale, textiffemale, name)
end

function M.female(textiffemale, textifmale, name)
    return M.isfemale(textiffemale, textifmale, name)
end

function M.waitress(name)
    return M.isfemale("waitress", "waiter", name)
end

function M.girlfriend(name)
    return M.isfemale("girlfriend", "boyfriend", name)
end

function M.femaleclothing(textiffemale, textifmale, name)
    name = name or "pc"
    return function(gamestate)
        local clothingstyle = gamestate.options.clothingstyle
        if clothingstyle == "genderbased" then
            return M.isfemale(textiffemale, textifmale, name)(gamestate)
        else
            return clothingstyle == "female" and textiffemale or textifmale
        end
    end
end

function M.brothersister(name)
    return M.female("sister", "brother", name)
end

function M.BrotherSister(name)
    return M.female("sister", "brother", name)
end

function M.sister(name)
    return M.brothersister(name)
end

function M.Sister(name)
    return M.BrotherSister(name)
end

function M.guysgirls(name)
    return M.isfemale("girls", "guys", name)
end

function M.GuysGirls(name)
    return M.isfemale("Girls", "Guys", name)
end

function M.girls(name)
    return M.guysgirls(name)
end

function M.Girls(name)
    return M.GuysGirls(name)
end

function M.underwearpanties(name)
    return M.gender("underwear", "panties", name)
end

function M.underwearpanty(name)
    return M.gender("underwear", "panty", name)
end

--[[
-- A function M.that takes in the name of a character whose printed name should
-- be displayed, and
-- returns a function M.that takes the gamestate and returns the name of the
-- specified character.
--
-- name: The system name of the character whose printed name shoul be displayed,
-- defaults to "pc"
--
-- returns A function that takes a gamestate, and returns the printed name of
-- the specified character
--]]
function M.name(name)
    name = name or "pc"
    log.trace("Building name dynamic text: ", name)
    return function(gamestate)
        return getcharacter(gamestate, name).printedname
    end
end

function M.names(name)
    name = name or "pc"
    return function(gamestate)
        local name = getcharacter(gamestate, name).printedname
        return name:sub(1, #name) == "s" and (name .. "'") or (name .. "'s")
    end
end

function M.hisher(name)
    return M.gender("his", "her", name)
end

function M.HisHer(name)
    return M.gender("His", "Her", name)
end

function M.her(name)
    return M.hisher(name)
end

function M.Her(name)
    return M.HisHer(name)
end

function M.hishers(name)
    return M.gender("his", "hers", name)
end

function M.hers(name)
    return M.hishers(name)
end

function M.HisHers(name)
    return M.gender("His", "Hers", name)
end

function M.Hers(name)
    return M.HisHers(name)
end

function M.heshe(name)
    return M.gender("he", "she", name)
end

function M.she(name)
    return M.heshe(name)
end

function M.HeShe(name)
    return M.gender("He", "She", name)
end

function M.She(name)
    return M.HeShe(name)
end

function M.himher(name)
    return M.gender("him", "her", name)
end

function M.HimHer(name)
    return M.gender("Him", "Her", name)
end

function M.herm(name)
    return M.himher(name)
end

function M.Herm(name)
    return M.HimHer(name)
end

function M.boygirl(name)
    return M.gender("boy", "girl", name)
end

function M.girl(name)
    return M.gender("boy", "girl", name)
end

function M.BoyGirl(name)
    return M.gender("Boy", "Girl", name)
end

function M.Girl(name)
    return M.BoyGirl(name)
end

function M.manwoman(name)
    return M.gender("man", "woman", name)
end


function M.menwomen(name)
    return M.gender("men", "women", name)
end

function M.women(name)
    return M.menwomen(name)
end

function M.woman(name)
    return M.manwoman(name)
end

function M.bastardbitch(name)
    return M.gender("asshole", "bitch", name)
end

function M.bitch(name)
    return M.bastardbitch(name)
end

function M.BastardBitch(name)
    return M.gender("Asshole", "Bitch", name)
end

function M.Bitch(name)
    return M.BastardBitch(name)
end

function M.himselfherself(name)
    return M.gender("himself", "herself", name)
end

function M.herself(name)
    return M.himselfherself(name)
end

function M.manlady(name)
    return M.gender("man", "lady", name)
end

function M.lady(name)
    return M.manlady(name)
end

function M.ManLady(name)
    return M.gender("Man", "Lady", name)
end

function M.Lady(name)
    return M.ManLady(name)
end

function M.sirmaam(name)
    return M.gender("sir", "ma'am", name)
end

function M.maam(name)
    return M.sirmaam(name)
end

function M.SirMaam(name)
    return M.gender("Sir", "Ma'am", name)
end

function M.Maam(name)
    return M.SirMaam(name)
end

function M.dudechick(name)
    return M.gender("dude", "chick", name)
end

function M.chick(name)
    return M.dudechick(name)
end

function M.DudeChick(name)
    return M.gender("Dude", "Chick", name)
end

function M.Chick(name)
    return M.DudeChick(name)
end

function M.guygal(name)
    return M.gender("guy", "gal", name)
end

function M.gal(name)
    return M.guygal(name)
end

function M.GuyGal(name)
    return M.gender("Guy", "Gal", name)
end

function M.Gal(name)
    return M.GuyGal(name)
end

function M.MrMiss(name)
    return M.gender("Mr", "Miss", name)
end

function M.Miss(name)
    return M.MrMiss(name)
end

function M.sexycostume(sexy, notsexy)
    return function(gamestate)
        return gamestate.keywords.sexyOutfit and sexy or notsexy
    end
end

function M.heels()
    return function(gamestate)
        local pc = gamestate.characters.pc
        return gamestate.keywords.sexyOutfit 
            and pc.attributes.gender == "female" 
            and "heels" 
            or "boots"
    end
end

function M.sexyCostume(sexy, notsexy)
  return M.sexycostume(sexy, notsexy)
end


function M.heroheroine(name)
    return M.gender("hero", "heroine", name)
end

function M.HeroHeroine(name)
    return M.gender("Hero", "Heroine", name)
end

function M.heroine(name)
    return M.heroheroine(name)
end

function M.Heroine(name)
    return M.HeroHeroine(name)
end

function M.superheroine(name)
    local heroheroinefunction = M.heroheroine(name)
    return function(gamestate)
        return "super" .. heroheroinefunction(gamestate)
    end
end

function M.Superheroine(name)
    return function(gamestate)
        return "Super" .. M.heroheroine(name)(gamestate)
    end
end

function M.superheroheroine(name)
    return M.superheroine(name)
end

function M.Superheroheroine(name)
    return M.Superheroine(name)
end

function M.superheroines(name)
    return function(gamestate)
        local text = M.superheroheroine(name)(gamestate)
        if text == "superhero" then
            return text .. "es"
        end
        return text .. "s"
    end
end

function M.Superheroines(name)
    return function(gamestate)
        return M.Superheroine(name)(gamestate) .. "s"
    end
end

function M.superheroheroines(name)
    return M.superheroines(name)
end

function M.Superheroheroines(name)
    return M.Superheroines(name)
end

function M.heroines(name)
    return function(combatstate)
        local heroheroine = M.heroheroine(name)(combatstate)
        return heroheroine .. (heroheroine == "hero" and "es" or "s")
    end
end

function M.Heroines(name)
    return function(combatstate)
        return M.HeroHeroine(name)(combatstate) .. "s"
    end
end

function M.hero(name)
    return M.heroheroine(name)
end

function M.Hero(name)
    return M.HeroHeroine(name)
end

function M.heroes(name)
    return M.heroheroine(name) .. "s"
end

function M.Heroes(name)
    return M.HeroHeroine(name) .. "s"
end

function M.height(short, average, tall, gargantuan, name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local characterheight = character.attributes.height
        return characterheight == "short"
            and short
            or characterheight == "average"
            and average
            or characterheight == "tall"
            and tall
            or characterheight == "gargantuan"
            and gargantuan
    end
end

function M.skincolor(name)
    name = name or "pc"
    return function(gamestate)
        return getcharacter(gamestate, name).attributes.skincolor
    end
end

function M.clothing(name, slot)
    name = name or "pc"
    return function(gamestate)
        return getcharacter(gamestate, name).clothing[slot].name
    end
end

function M.wearing(outfit, ifwearingoutfit, ifnot, name)
    name = name or "pc"
    return function(gamestate)
        local clothing = getcharacter(gamestate, name).clothing
        for _, slot in pairs(clothing._keys) do
            if clothing[slot].name == outfit then
                return ifwearingoutfit
            end
        end
        return ifnot
    end
end

function M.shirt(name)
    return M.clothing(name, "upper")
end

function M.pants(name)
    return M.clothing(name, "lower")
end

function M.panties(name)
    return M.clothing(name, "underwear")
end

function M.underwear(name)
    return M.panties(name)
end

function M.clothingslot(name, slot, outfit, textifoutfit, textotherwise)
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        return character.clothing[slot].name == outfit
            and textifoutfit
            or textotherwise
    end
end

function M.upper(name, outfit, textifoutfit, textotherwise)
    return M.clothingslot(name, "upper", outfit, textifoutfit, textotherwise)
end

function M.lower(name, outfit, textifoutfit, textotherwise)
    return M.clothingslot(name, "lower", outfit, textifoutfit, textotherwise)
end

function M.Pants(name)
    return function(gamestate)
        return string.gsub(M.pants(name)(gamestate), "^%l", string.upper)
    end
end

function M.baring(baringtext, modesttext, name)
    name = name or "pc"
    modesttext = modesttext or ""
    return function(gamestate)
        return getcharacter(gamestate, name).clothing.underwear.risque >= 5
            and baringtext
            or modesttext
    end
end

local MUSCULATURE_ADJECTIVES = {
    soft={"rippling", "jumping", "bobbling", "shaking", "wobbling"},
    fit={
        "rippling",
        "spasming",
        "bouncing",
        "shaking",
        "bobbing",
        "oscillating",
        "quivering"
    },
    muscular={"quivering", "bobbing", "shuddering", "shifting"}
}

local lastadjective = {}

function M.bodyadj(name, attribute, adjectivetable)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local adjectives = adjectivetable[character.attributes[attribute]]
        -- Saves us from building a potentially expensive string every
        -- time we use bodyadj.
        if not adjectives then
            assert(
                adjectives,
                "Unknown attribute" .. attribute .. ": "
                .. "'" .. (character.attributes[attribute]  or "nil") .. "'"
                .. " for character: "
                .. scarlettables.string(character)
            )
        end
        if type(adjectives) ~= "table" then
            assert(
                false, 
                "Expected a table, but got '" .. scarlettables.string(adjectives) 
                        .. "'"
            )
        end
        local adjective
        repeat
            adjective = adjectives[math.random(#adjectives)]
        until (#adjectives <= 1 or adjective ~= lastadjective[attribute])
        lastadjective[attribute] = adjective
        return adjective
    end
end

local MUSCULATURE_VERBS = {
    soft={"rippled", "jumped", "bobbled", "shook", "wobbled"},
    fit={
        "rippled",
        "spasmed",
        "bounced",
        "shook",
        "bobbed",
        "oscillated",
        "quivered"
    },
    muscular={"quivered", "bobbed", "shuddered"}
}

local BUM_SIZE = {
    slim="petite",
    slender="heart-shaped",
    voluptuous="big",
    heavyset="huge"
}

function M.bumsize(name)
    return M.bodyadj(name, "bodytype", BUM_SIZE)
end

function M.big(name)
    return M.bumsize(name)
end

function M.muscleadj(name)
    return M.bodyadj(name, "musculature", MUSCULATURE_ADJECTIVES)
end

function M.muscleverb(name)
    return M.bodyadj(name, "musculature", MUSCULATURE_VERBS)
end

function M.muscled(name)
    return M.muscleverb(name)
end

function M.muscleing(name)
    return M.muscleadj(name)
end

function M.soft(name)
    return M.bodyadj(
        name,
        "musculature",
        {soft={"soft"}, fit={"firm"}, muscular={"hard"}}
    )
end

function M.rippled(name)
    return M.muscleverb(name)
end

function M.rippling(name) 
    return M.muscleadj(name)
end

function M.wobbling(name)
    return M.rippling(name)
end

function M.muscle(softtext, fittext, musculartext, name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local musculature = character.attributes.musculature
        return musculature == "soft" and softtext
            or musculature == "fit" and fittext
            or musculature == "muscular" and musculartext
    end
end

function M.muscular(softtext, fittext, musculartext, name)
    return M.muscle(softtext, fittext, musculartext, name)
end

function M.musculature(softtext, fittext, musculartext, name)
    return M.muscle(softtext, fittext, musculartext, name)
end

function M.wobbled(name)
    return M.muscle("wobbled", "jiggled", "rippled")
end

function M.wobbling(name)
    return M.muscle("wobbling", "jiggling", "rippling")
end

local BODYTYPE_ADJECTIVES = {
    slim={
        "small",
        "petite",
        "heart-shaped",
        "lithe",
        "svelt",
        "compact",
        "narrow",
        "willowy"
    },
    slender={"plump", "round", "curved", "svelt", "compact"},
    voluptuous={
        "ample",
        "curvaceous",
        "protruding",
        "shapely",
        "curvy",
        "rotund"
    },
    heavyset={
        "fleshy",
        "wide",
        "expansive",
        "corpulent",
        "sizeable",
        "generous",
        "huge"
    }
}

function M.bumadj(name)
    return M.bodyadj(name, "bodytype", BODYTYPE_ADJECTIVES)
end

local BREAST_ADJECTIVES = {
    slim={"pert"},
    slender={"perky"},
    voluptuous={"ample"},
    heavyset={"voluminous"}
}
function M.ample()
    return M.bodyadj(name, "bodytype", BREAST_ADJECTIVES)
end

function M.rotund(name)
    return M.bumadj(name)
end

function M.bodytype(slim, slender, voluptuous, heavyset, name)
    name = name or "pc"
    assert(slim, "Missing text for slim in bodytype.")
    assert(slender, "Missing text for slender in bodytype.")
    assert(voluptuous, "Missing text for voluptuous in bodytype.")
    assert(heavyset, "Missing text for heavyset in bodytype.")
    return function(gamestate)
        local bodytype = getcharacter(gamestate, name).attributes.bodytype
        local result =
            bodytype == "slim" and slim or
            bodytype == "slender" and slender or
            bodytype == "voluptuous" and voluptuous or
            bodytype == "heavyset" and heavyset
        return assert(result, "Unknown body type: " .. bodytype)
    end
end

function M.bumsize(name)
    local bumsizetext = M.bodytype("pert", "heart-shaped", "big", "massive", name)
    return function(gamestate)
        return bumsizetext(gamestate)
    end
end

local HEIGHT_RANK = {
    short=1,
    average=2,
    tall=3,
    gargantuan=4
}

function M.height(short, average, tall, gargantuan, name)
    name = name or "pc"
    return function(gamestate)
        local height = getcharacter(gamestate, name).attributes.height
        local result =
            height == "short" and short
            or height == "average" and average
            or height == "tall" and tall
            or height == "gargantuan" and gargantuan
        return assert(result, "Unknown height: " ..  height)
    end
end

function M.towers(towerername, towereename)
    assert(
        towerername and towereename,
        "Both someone to tower, and someone to be towered over must be "
        .. "provided."
    )
    return function(gamestate)
        local towerer = getcharacter(gamestate, towerername)
        local toweree = getcharacter(gamestate, towereename)
        local towererrank = HEIGHT_RANK[towerer.attributes.height]
        local towereerank = HEIGHT_RANK[toweree.attributes.height]
        local heightdifference = towererrank - towereerank
        return
            heightdifference >= 2 and "towered over"
            or heightdifference >= 1 and "loomed over"
            or heightdifference == 0 and "glared at"
            or heightdifference < 0 and "glared up at"
    end
end


function M.pantsskirt(pantstext, skirttext, name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local bottomstype = character.clothing.lower.type
        if bottomstype == "skirt" or bottomstype == "dress" then
          return skirttext or "skirt"
        end
        return pantstext or "pants"
    end
end

function M.costume(sexy, notsexy, otherwise, name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local clothingtype = character.clothing.underwear.type
        if clothingtype == "costume" then
            return M.sexycostume(sexy, notsexy)(gamestate)
        else
            return otherwise
        end
    end
end

function M.pantsdress(pantstext, dresstext, name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local bottomstype = character.clothing.lower.type
        if bottomstype == "dress" then
          return dresstext
        end
        return pantstext
    end
end

function M.skirt(skirttext, pantstext, name)
    return M.pantsskirt(pantstext, skirttext, name)
end

function M.up(name)
    return M.skirt("up", "down", name)
end

function M.down(name)
    return M.up(name)
end

function M.raised(name)
    return M.skirt( "raised", "lowered", name)
end

function M.lowered(name)
    return M.raised(name)
end

function M.sore(sore, notsore, name)
    name = name or "pc"
    return function(gamestate)
        return getcharacter(gamestate, name).bummarks and sore or notsore
    end
end

function M.bum(sore, notsore, name)
    return M.sore(sore, notsore, name)
end

function M.thong(thong, nothong, name)
    name = name or "pc"
    return function(gamestate)
        local underwear = getcharacter(gamestate, name).clothing.underwear
        return underwear.type == "thong" and thong or nothong
    end
end

function M.keyword(keyword, keywordtext, nokeywordtext, placeholder)
    nokeywordtext = nokeywordtext or ""
    placeholder = placeholder or nokeywordtext
    return function(gamestate)
        return gamestate.keywords.placeholder
            and placeholder
            or gamestate.keywords[keyword]
            and keywordtext
            or nokeywordtext
    end
end

function M.risque(risquetext, modesttext, threshhold, name)
    name = name or "pc"
    threshhold = threshhold or 5
    return function(gamestate)
        local clothing =  getcharacter(gamestate, name).clothing
        local lowerclothing = clothing.lower or clothing.underwear
        return lowerclothing.risque >= threshhold and risquetext or modesttext
    end
end

function M.cheeky(cheekytext, modesttext, name)
    return M.risque(cheekytext, modesttext, name)
end

function M.spankable(spankable, notspankable)
    return function(gamestate)
        return gamestate.options.spankableplayer == "yes" and spankable or notspankable
    end
end

function M.pantstype(name)
    name = name or "pc"
    return function(gamestate)
        local clothing = getcharacter(gamestate, name).clothing
        return (clothing.lower or clothing.underwear).type
    end
end

function M.Pantstype(name)
  name = name or "pc"
  return function(gamestate)
    local type = M.pantstype(name)(gamestate)
    return string.gsub(type, "^%l", string.upper)
  end
end

function M.hairstyle(name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local style = character.attributes.hairstyle
        return style == "down" and "hair" or style
    end
end

function M.hair(down, pigtails, ponytail, frenchbraid, buns, mohawk, name)
    name = name or "pc"
    local styletext = {
        down=down,
        ["pig tails"]=pigtails,
        ponytail=ponytail,
        ["french braid"]=frenchbraid,
        buns=buns,
        mohawk=mohawk
    }
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local hairstyle = character.attributes.hairstyle
        local text = assert(
            styletext[hairstyle],
            "Missing text for hairstyle " .. "'" .. hairstyle .. "'"
        )
        return text
    end
end

function M.hairlength(name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local length = character.attributes.hairstyle
    end
end

function M.hairlengthmsg(short, shoulderlength, backlength, name)
    name = name or "pc"
    local lengthtext = {
        short=short,
        shoulderlength=shoulderlength,
        backlength=backlength
    }
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local hairlength = character.attributes.hairlength
        local text = assert(
            lengthtext[hairlength],
            "Missing text for hairlength " .. "'" .. hairlength .. "'"
        )
        return text
    end
end

function M.haircolor(name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        return character.attributes.haircolor
    end
end

function M.assholecunt(name)
    return M.gender("asshole", "cunt", name)
end

function M.cunt(name)
    return M.gender("asshole", "cunt", name)
end

function M.pussy(name)
    return M.gender("cock", "pussy", name)
end

local SPANKER = 1
local SPANKEE = -1
local SWITCH = 0
local function spankerspankeeswitch(character)
    local spanker = character.statistics.spanker or 1
    local spankee = character.statistics.spankee or 1
    if spanker == spankee then
        return SWITCH
    end
    return spanker >= 2 * spankee
        and SPANKER
        or spankee >= 2 * spanker
        and SPANKEE
        or SWITCH
end

function M.role(ifspanker, ifspankee, ifswitch, name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local role = spankerspankeeswitch(character)
        return role == SWITCH
            and ifswitch
            or role == SPANKER
            and ifspanker
            or ifspankee
    end
end

function M.spankee(ifspankee, ifswitch, ifspanker, name)
    return M.role(ifspanker, ifspankee, ifswitch, name)
end

local COMPETENT = 1
local INCOMPETENT = -1
local BALANCED = 0
local function competentincompetentbalanced(character)
    local victory = character.statistics.victory
    local defeat = character.statistics.defeat
    -- Handles the "both are zero" edge case.
    if victory == defeat then
        return BALANCED
    end
    return victory >= 2 * defeat
        and COMPETENT
        or defeat >= 2 * victory
        and INCOMPETENT
        or BALANCED
end


function M.competence(ifcompetent, ifincompetent, ifbalanced, name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local competence = competentincompetentbalanced(character)
        return competence == BALANCED
            and ifbalanced
            or competence == COMPETENT
            and ifcompetent
            or ifincompetent
    end
end

--[[
-- Prints the given text based on which choice the
-- the player made last. 
--
-- This function assumes up 
-- to three choices provided, but that's fine. This
-- will mostly to allow a bit of variation based on
-- roleplaying choices the player made.
--]]
function M.choice(path1, path2, path3)
    local pathtext = {path1, path2, path3}
    return function(gamestate)
        return assert(
            pathtext[tonumber(gamestate.chosenPath)],
            "No text at index " .. (gamestate.chosenPath or "nil")
        )
    end
end

return M
